package geniusweb.protocol.session.mopac;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;

import geniusweb.actions.Action;
import geniusweb.actions.PartyId;
import geniusweb.inform.Agreements;
import geniusweb.inform.OptIn;
import geniusweb.progress.Progress;
import geniusweb.progress.ProgressRounds;
import geniusweb.protocol.ProtocolException;
import geniusweb.protocol.partyconnection.ProtocolToPartyConn;
import geniusweb.protocol.partyconnection.ProtocolToPartyConnections;
import geniusweb.protocol.session.SessionResult;
import geniusweb.protocol.session.SessionState;
import geniusweb.protocol.session.mopac.phase.OfferPhase;
import geniusweb.protocol.session.mopac.phase.OptInPhase;
import geniusweb.protocol.session.mopac.phase.Phase;
import geniusweb.protocol.session.saop.SAOPSettings;
import geniusweb.references.PartyWithProfile;
import geniusweb.voting.VotingEvaluator;

public class MOPACState implements SessionState {

	private final Phase phase;
	private final MOPACSettings settings;
	private final Map<PartyId, PartyWithProfile> partyprofiles;
	private final ProtocolToPartyConnections connections;
	private final List<Action> actions;
	private final Progress progress;

	/**
	 * Creates the initial state from the given settings and progress=null
	 * 
	 * @param settings the {@link SAOPSettings}
	 */
	public MOPACState(MOPACSettings settings) {
		this(null, Arrays.asList(),
				new ProtocolToPartyConnections(Collections.emptyList()), null,
				settings, Collections.emptyMap());

	}

	/**
	 * @param phase         The Phase, or null if still initializing. Phase is
	 *                      set something only after we have connections.
	 * @param actions       the legal actions that have been done in the
	 *                      negotiation. A new-sub-list is added for each phase
	 *                      change. It is assumed that the last element of the
	 *                      list is the list with actions taken in the current
	 *                      phase. first action in first list is the oldest.
	 *                      This MUST NOT contain illegal actions. Parties doing
	 *                      illegal actions are killed and the offending action
	 *                      ends up in the stacktrace. Previous actions of
	 *                      crashed parties remain standing and valid.
	 * @param conns         the existing party connections. we assume ownership
	 *                      of this so it should not be modified although
	 *                      connections may of course break.
	 * @param progress      the {@link Progress} line. can be null if not yet
	 *                      known. This happens because during setup we first
	 *                      add all connections, and only when we have that we
	 *                      set the correct deadline.
	 * @param settings      the {@link SAOPSettings}
	 * @param partyprofiles map with the {@link PartyWithProfile} for connected
	 *                      parties. null is equivalent to an empty map.
	 * @param evaluator     the {@link VotingEvaluator}
	 */
	@JsonCreator
	protected MOPACState(@JsonProperty("phase") Phase phase,
			@JsonProperty("actions") List<Action> actions,
			@JsonProperty("connections") ProtocolToPartyConnections conns,
			@JsonProperty("progress") Progress progress,
			@JsonProperty("settings") MOPACSettings settings,
			@JsonProperty("partyprofiles") Map<PartyId, PartyWithProfile> partyprofiles) {
		this.phase = phase;
		this.actions = actions;
		this.connections = conns;
		this.progress = progress;
		this.settings = settings;
		this.partyprofiles = partyprofiles;
	}

	/**
	 * Can be called after all connections are initialized.
	 * 
	 * @param now current time ms since 1970
	 * 
	 * @return state with the initial partystates set.
	 */
	public MOPACState initPhase(long now) {
		PartyStates partyStates = new PartyStates(getPowers());
		Phase firstPhase = new OfferPhase(null, partyStates,
				now + getAvailablePhaseTime(now),
				settings.getVotingEvaluation());
		return new MOPACState(firstPhase, actions, connections, progress,
				settings, partyprofiles);
	}

	@Override
	public List<Action> getActions() {
		return Collections.unmodifiableList(actions);
	}

	@Override
	public Progress getProgress() {
		return progress;
	}

	public Map<PartyId, PartyWithProfile> getPartyProfiles() {
		return Collections.unmodifiableMap(partyprofiles);
	}

	/**
	 * @return all current/remaining active party connections. Finished/crashed
	 *         parties should be removed immediately.
	 */
	public ProtocolToPartyConnections getConnections() {
		return connections;
	}

	/**
	 * @param now current time ms since 1970
	 * @return the max possible duration in ms of the NEXT phase. Maybe 0 or
	 *         negative if past deadline.
	 */
	private Long getAvailablePhaseTime(long now) {
		// explicit check, to check also the round counts.
		if (incrementProgress().isPastDeadline(now + Phase.PHASE_MINTIME))
			return 0l;
		return Math.min(progress.getTerminationTime().getTime() - now,
				Phase.PHASE_MAXTIME);
	}

	/**
	 * 
	 * @param connection   the new {@link ProtocolToPartyConn}
	 * @param partyprofile the {@link PartyWithProfile} that is associated with
	 *                     this state
	 * @return new {@link MOPACState} with the new connection added. This call
	 *         ignores the progress (does not check isFinal) because we uses
	 *         this during the setup where the deadline is not yet relevant.
	 */
	protected MOPACState with(ProtocolToPartyConn connection,
			PartyWithProfile partyprofile) {
		if (phase != null)
			throw new IllegalStateException(
					"Adding connections only allowed while initializing");

		ProtocolToPartyConnections newconns = getConnections().with(connection);
		Map<PartyId, PartyWithProfile> newprofiles = new HashMap<>(
				partyprofiles);
		newprofiles.put(connection.getParty(), partyprofile);
		return new MOPACState(null, actions, newconns, progress, settings,
				newprofiles);
	}

	/**
	 * @param id the {@link PartyId} of the party that failed.
	 * @param e  the {@link ProtocolException} that occurred in the party
	 * @return a new state with the error set.
	 */
	public MOPACState with(ProtocolException e) {
		return new MOPACState(phase.with(e), actions, connections, progress,
				settings, partyprofiles);
	}

	/**
	 * Sets the progress for this session. Can be set only if progress=null.
	 * Should be set in INIT phase.
	 * 
	 * @param newprogress the new progress
	 * @return new {@link MOPACState} with the progress set
	 */
	public MOPACState with(Progress newprogress) {
		if (progress != null || newprogress == null || phase != null) {
			throw new IllegalArgumentException(
					"progress must be null, newprogress must be not null and phase must be INIT");
		}
		return new MOPACState(phase, actions, connections, newprogress,
				getSettings(), partyprofiles);
	}

	/**
	 * Start the next phase. If new phase is OfferPhase, we increase progress.
	 * Actions is reset to empty. does nothing if not
	 * {@link #isNewPhasePossible(long)}
	 * 
	 * @param now current time
	 * @return new {@link MOPACState} with phase initialized for next phase.
	 */
	public MOPACState nextPhase(long now) {
		long remainingNegoTime = progress.getTerminationTime().getTime() - now;
		Phase newphase = phase.next(now,
				Math.min(remainingNegoTime, Phase.PHASE_MAXTIME));

		return new MOPACState(newphase, actions, connections,
				incrementProgress(), getSettings(), partyprofiles);
	}

	/**
	 * 
	 * @return the next progress. Progress round advances if phase is
	 *         {@link OptIn}.
	 */
	private Progress incrementProgress() {
		if (progress instanceof ProgressRounds && phase instanceof OptInPhase)
			return ((ProgressRounds) progress).advance();
		return progress;
	}

	@Override
	public boolean isFinal(long now) {
		return phase != null && phase.isFinal(now) && !isNewPhasePossible(now);
	}

	/**
	 * When this is called, all parties should have acted.
	 * 
	 * @param now
	 * @return
	 */
	public boolean isNewPhasePossible(long now) {
		// System.out.println("phase=" + phase);
		boolean a = phase.getPartyStates().getNegotiatingParties().size() >= 2
				&& getAvailablePhaseTime(now) > Phase.PHASE_MINTIME;
		// System.out.println("newstate possible:" + a);
		return a;
	}

	@Override
	public Agreements getAgreements() {
		return phase.getPartyStates().getAgreements();
	}

	@Override
	public MOPACSettings getSettings() {
		return settings;
	}

	public Phase getPhase() {
		return phase;
	}

	/**
	 * Check if action is allowed. Add action to the list of actions. Notice,
	 * this does NOT check if we need to step to the next phase, because
	 * deciding that is also depending on time-outs.
	 * 
	 * @param actor  the actor that did this action. Can be used to check if
	 *               action is valid. NOTICE caller has to make sure the current
	 *               state is not final. MUST NOT be null.
	 * @param action the action that was proposed by actor. MUST NOT be null.
	 * @param now    the current time in ms since 1970, see
	 *               {@link System#currentTimeMillis()}
	 */

	public MOPACState with(PartyId actor, Action action, long now) {
		return new MOPACState(phase.with(actor, action, now), actions,
				connections, progress, settings, partyprofiles);
	}

	@Override
	public SessionResult getResult() {
		List<Double> emptylist = new LinkedList<>();
		for (int n = 0; n < getConnections().size(); n++)
			emptylist.add(0d);
		return new SessionResult(getSettings().getAllParties(), getAgreements(),
				emptylist, null);
	}

	/**
	 * @return the power of all parties as set in their parameters, default =1.
	 *         bad power values (non-integer, or <1) are ignored.
	 */
	private Map<PartyId, Integer> getPowers() {
		Map<PartyId, Integer> map = new HashMap<>();
		for (PartyId pid : partyprofiles.keySet()) {
			Object power = partyprofiles.get(pid).getParty().getParameters()
					.get("power");
			if (power == null || !(power instanceof Integer)
					|| ((Integer) power) < 1)
				power = 1;
			map.put(pid, (Integer) power);
		}
		return map;
	}

	@Override
	public String toString() {
		return "MOPACState[" + phase + "," + settings + "," + partyprofiles
				+ "," + connections + "," + progress + "," + "]";
	}

	/**
	 * 
	 * @return a wrapped-up state, with all parties doen an action or kicked,
	 *         and agreements collected
	 */
	public MOPACState finishPhase() {
		Phase newphase = phase.finish();
		LinkedList<Action> newactions = new LinkedList<Action>(actions);
		newactions.addAll(newphase.getPartyStates().getActions());
		return new MOPACState(newphase, newactions, connections, progress,
				settings, partyprofiles);
	}

}
